/*******************************************************************************
* sirf.cpp <=> sirf.h: NMEA 0183 & SiRF Binary Protocol library for Arduino    *
* Copyright (c) 2012 Jacob Dilles <jdilles@gmu.edu>        All right reserved. *
*                                                                              *
* This software has been released under The MIT License (MIT):                 *
*                                                                              *
* Permission is hereby granted, free of charge, to any person obtaining a copy *
* of this software and associated documentation files (the "Software"), to     *
* deal in the Software without restriction, including without limitation the   *
* rights to use, copy, modify, merge, publish, distribute, sublicense, and/or  *
* sell copies of the Software, and to permit persons to whom the Software is   *
* furnished to do so, subject to the following conditions:                     *
*                                                                              *
* (1) The above copyright notice and this permission notice shall be included  *
* in all copies or substantial portions of the Software.                       *
*                                                                              *
* (2) THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  *
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       *
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING      *
* FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS *
* IN THE SOFTWARE.                                                             *
*                                                                              *
* Notice: The author has no affiliation with SiRF Technology, Inc. or CSR plc. *
*                                                                              *
* Many thanks to Maarten Lamers for his NMEA library:                          *
*       http://www.maartenlamers.com/nmea/                                     *
* Which inspired this project and got it rolling in the right direction.       *
********************************************************************************/    

#include "sirf.h"
#include <stdlib.h>
#include <string.h>
#include <stdint.h>


#define HI(i)    ( 0xFF & (((unsigned int)i) >> 8) ) 
#define LO(i)    ( 0xFF & ((unsigned int)i) ) 
#define IV(i)    ( 0xFF & ((unsigned int)i) ) 
#define times10plus(x, y)  ((x<<3) + (x<<1) + y)
#define HEX_NIB_L(c)     ( _BIN2HEX[0x0F & c] );
#define HEX_NIB_H(c)     ( _BIN2HEX[0x0F & (c>>4)]);

/* File-private functions */
static char _bcd2dec2(char** sp);
static char _dehex(char c);
static fp_t _lat2dd(char* s);
static fp_t _lon2dd(char* s);

/* File-private variables */
static const char _BIN2HEX[] = "0123456789ABCDEF";


/* Macros */
#define CLEAR_PTR_ST_GPRMC(d)    d->status = 'V'; d->utc_year = 0; d->utc_month = 0; d->utc_day = 0; d->utc_hour = 0; d->utc_min = 0; d->utc_sec = 0; d->lat = NAN; d->lon = NAN; d->speed = NAN; d->angle = NAN;

#define putGPSchar(x)    _serial->write((unsigned char)x)
#define getGPSchar()     _serial->read()
#define gpsAvailable()   _serial->available()

static char _bcd2dec2(char** sp);
static char _dehex(char c);

static fp_t _lat2dd(char* s);
static fp_t _lon2dd(char* s);

SiRF::SiRF(void){
      
    _tc = 0;       /* term  # counter */
    _cc = 0;       /* term character counter */
    _sc = 0;       /* sentence counter */
    _bc = 0;       /* binary counter */
    _checksum = 0; /* checksum */ 
    _mode = NMEA;     /* binary or NMEA */
    _state = RESET_FSM;
    _serial = NULL;
    _baud = 0;
}


/* Object-Oriented Functions */
void SiRF::begin( const HardwareSerial * stream, unsigned long baud)
{
    _serial = (const_cast<HardwareSerial *>(stream));
    _baud = (uint16_t)baud;
    _serial->begin(_baud);

}

mode_t SiRF::receive(void){
   while(_serial->available()) 
      if(_decode(_serial->read())) return _mode;
   return INVALID;
}

int SiRF::parseGPRMC(struct st_gprmc * d)
{
    char * p;

    if( !(_break_sent()) ){
        /* Not in NMEA mode */
        return FALSE;
    }
    
    if( strncmp("GPRMC", _term[0], 6) ){
        /* Not a correct term */
        return FALSE;
    }
    
    CLEAR_PTR_ST_GPRMC(d);

    p = _term[2];   /* status */
    if(p){
      d->status = *p;
    }
        
    p = _term[1];   /* UTC time */
    if(p){
        d->utc_hour = _bcd2dec2(&p);
        d->utc_min = _bcd2dec2(&p);
        d->utc_sec = _bcd2dec2(&p);
    }
    
    p = _term[9];    /* UTC date */
    if(p){
        d->utc_day = _bcd2dec2(&p);
        d->utc_month = _bcd2dec2(&p);
        d->utc_year = _bcd2dec2(&p);
    }  
    
    p = _term[3];   /* Latitude */
    if(p) {
        fp_t lat = _lat2dd(p);

        if( (p = _term[4]) ) /* Hemisphere */
        {
                 if( *p == 'N' ) d->lat = lat;
            else if (*p == 'S' ) d->lat = -lat;
        }
            /* else ignore field  */
    }
    
    p = _term[5];   /* Longitude */
    if(p) {
        fp_t lon = _lon2dd(p);
        if( (p = _term[6]) ) /* Hemisphere */
         {
                 if( *p == 'E' ) d->lon = lon;
            else if (*p == 'W' ) d->lon = -lon;
         }
        /* else ignore field  */
    }
    
    p = _term[7]; /* Speed */
    if(p)  d->speed = atof(p);
    
    p = _term[8]; /* Bearing */
    if(p)  d->angle = atof(p);
    
    return TRUE;
}

/* 
 Sends a NMEA String, calculating the checksum as necessary
    params:     s - pointer to a string
    returns:    0 on sucess, nz on failure
Reference: http://rietman.wordpress.com/2008/09/25/how-to-calculate-the-nmea-checksum/ 
*/
int SiRF::sendNMEA(char* s)
{
    /* If we're in binary mode, don't bother trying... */
    if(_mode != NMEA) {
        /* but set the mode to NMEA so if they really want to they can send 
        on the next go */
        _mode = NMEA;
        return 1;
    }
    /* maybe the user copied a whole string. Ignore... */
    if( (*s) == '$' ) *s++;
    
    unsigned char checksum, c;
    
    /* write the start character */
    c = '$';                                putGPSchar(c);
    
    /* write the payload, and calculate the checksum */
    checksum = 0;
    while( (*s) )
    {
        if( (c = (*s++)) == '*' ) {break;}  putGPSchar(c);
        checksum ^= c;
    }
    
    /* write checksum */
    c = '*'; putGPSchar(c);
    c = HEX_NIB_H(checksum);                putGPSchar(c);
    c = HEX_NIB_L(checksum);                putGPSchar(c);
    
    /* CR/LF */
    c = 0x0D;                               putGPSchar(c);
    c = 0x0A;                               putGPSchar(c);
    return 0;
}
int SiRF::sendNMEA(const char* s)
{
    return sendNMEA(const_cast<char *> (s));
}
int SiRF::sendNMEA(const __FlashStringHelper *ifsh)
{
    /* If we're in binary mode, don't bother trying... */
    if(_mode != NMEA) {
        /* but set the mode to NMEA so if they really want to they can send 
        on the next go */
        _mode = NMEA;
        return 1;
    }
    const char PROGMEM *p = (const char PROGMEM *)ifsh;
    unsigned char checksum, c;
    
    /* maybe the user copied a whole string. Ignore... */
    c = pgm_read_byte(p);
    if( c == '$' ) p++;
    
    /* write the start character */
    c = '$';                                putGPSchar(c);
    
    /* write the payload, and calculate the checksum */
    checksum = 0;
    while( 1 )
    {
        c = pgm_read_byte(p++);
        if( c == 0 || c == '*' ) break; 
        
        putGPSchar(c);
        
        checksum ^= c;
    }
    
    /* write checksum */
    c = '*'; putGPSchar(c);
    c = HEX_NIB_H(checksum);                putGPSchar(c);
    c = HEX_NIB_L(checksum);                putGPSchar(c);
    
    /* CR/LF */
    c = 0x0D;                               putGPSchar(c);
    c = 0x0A;                               putGPSchar(c);
    return 0;
}
int SiRF::sendNMEA(const String &s)
{
    /* If we're in binary mode, don't bother trying... */
    if(_mode != NMEA) {
        /* but set the mode to NMEA so if they really want to they can send 
        on the next go */
        _mode = NMEA;
        return 1;
    }
    unsigned char checksum, c;
    uint16_t i = 0;
    /* maybe the user copied a whole string. Ignore... */
    if( (c = s[i]) == '$' ) i++;
    
    /* write the start character */
    c = '$';                                putGPSchar(c);
    
    /* write the payload, and calculate the checksum */
    checksum = 0;
    while( 1 )
    {
        c = s[i++];
        if( c == 0 || c == '*' ) break; 
        
        putGPSchar(c);
        checksum ^= c;
    }
    
    /* write checksum */
    c = '*';                                putGPSchar(c);
    c = HEX_NIB_H(checksum);                putGPSchar(c);
    c = HEX_NIB_L(checksum);                putGPSchar(c);
    
    /* CR/LF */
    c = 0x0D;                               putGPSchar(c);
    c = 0x0A;                               putGPSchar(c);
    return 0;
}

/* 
    Sends an SiRF Binary Protocol String, calcuates checksums, etc.. 
    params:     buffp - pointer to buffer
                lenp - length to send
    returns:    boolean error
    reference: SiRF Binary Protocol Reference Manual Revision 2.4 November 2008 
*/
int SiRF::sendBIN(char* buffp, int16_t lenp)
{
    /* Not in binary mode, no point in sending */
    if(_mode != BINARY) {
        /* but set mode to binary in case they really want to send it 
          they can try again */
        _mode = BINARY; 
        return 1;
    }
    
    /* too big - limit = 2^10-1 */
    if(lenp &~ (0x3FF)) return 1;
    /* null pointer */
    if(!buffp)  return 1;
    
    uint16_t checksum ;
    unsigned char c;
    
    /* send start sequence */
    c = 0xA0;                               putGPSchar(c);
    c = 0xA2;                               putGPSchar(c);
    
    
    if(lenp)
    {
        /* variable length mode: send 16 bit payload length */
        c = HI(lenp);                       putGPSchar(c);
        c = LO(lenp);                       putGPSchar(c);
    }
    else 
    {
        /* fixed length mode, length encoded in data */
        c = *(buffp++);                     putGPSchar(c);
        lenp = (((uint16_t)c) << 8) & 0x3FF;
        c = *(buffp++);                     putGPSchar(c);
        lenp |= ((uint16_t)c);
    }
    
    /* send payload, and calculate checksum */
    checksum = 0;
    while(lenp--)
    {
        c = *(buffp++);                     putGPSchar(c);
        checksum += c;
    }
    
    /* send 16 bit checksum */
    c = HI(checksum);                       putGPSchar(c);
    c = LO(checksum);                       putGPSchar(c);
 
    /* send stop sequence */
    c = 0xB0;                               putGPSchar(c);
    c = 0xB3;                               putGPSchar(c);        

    return 0;
}
int SiRF::sendBIN(const char* buffp, int16_t lenp)
{
    return sendBIN(const_cast<char *> (buffp), lenp);
}
int SiRF::sendHEX(const char* str)
{
    unsigned char c, h, l;
    uint16_t len = strnlen(str, 0x3fE), checksum;
    
    if( len & 0x01) {
        return 1;
    }
    /* actual length is half of hex length */
    len >>= 1;

    /* send start sequence */
    c = 0xA0;                               putGPSchar(c);
    c = 0xA2;                               putGPSchar(c);

    /* send 16 bit payload length */
    c = HI(len);                            putGPSchar(c);
    c = LO(len);                            putGPSchar(c);
        
    /* send payload, and calculate checksum */
    checksum = 0;
    while( len-- ) {
        h = *(str++);
        l = *(str++);
        c = (_dehex(h)<<4) | (_dehex(l));   putGPSchar(c);
        checksum += c;
    }

    /* send 16 bit checksum */
    c = HI(checksum);                       putGPSchar(c);
    c = LO(checksum);                       putGPSchar(c);
 
    /* send stop sequence */
    c = 0xB0;                               putGPSchar(c);
    c = 0xB3;                               putGPSchar(c);        

    return 0;
}



char*  SiRF::term(int t)
{
    if(_state != NMEA_VALID) return NULL;
    if(t > _tc) return NULL;
    if(_sc) _break_sent();
    return _term[t];
}

char*   SiRF::sentence()
{
    if(_sc)
    switch(_state){
        case NMEA_VALID:
            return _sent;
        case BIN_VALID:
            return (_sent + 4);
        default:
            return NULL;
    }
#ifdef  _KEEP_SENTENCE_
    else switch(_state){
        case NMEA_VALID:
            return _sent2;
        case BIN_VALID:
            return (_sent2 + 4);
        default:
    }
#endif
     return NULL;
}

int     SiRF::sentenceLength()
{
    switch(_state){
        case NMEA_VALID:
        case BIN_VALID:
            return _bc;
        default:
            return 0;
    }
}




/**************************** Class-private Functions **********************/

/* 
 Decode a single character from the serial stream 
    params: c - character to parse
    returns: TRUE if we have reached the end of a valid string, FALSE otherwise
  */
int SiRF::_decode(char c)
{   

    if(_sc > MAX_SENTENCE_LEN || _tc > MAX_NUM_TERMS || _cc > MAX_TERM_LEN){
        _state = RESET_FSM;
    }

    if(_state == RESET_FSM || _state == NMEA_VALID || _state == BIN_VALID){
            /* Reset everything */
            _tc = 0;
            _cc = 0;
            _sc = 0;
            _bc = 0;
            _checksum = 0;
            
            /* wait for the second start character */
            _state = WAIT_START;
    }

    /* always store in buffer */
    _sent[_sc++] = c;
 
    /* Massive FSM of doom!! */
    switch(_state)
    {          
        case WAIT_START:
            /* Look for a valid start sequence */
           
            /* Do we have a valid binary start sequence? */
            if( (IV(_sent[0]) == 0xA0) && ( IV(c) == 0xA2) ) {
                /* got the binary mode start sequence */
                _mode = BINARY;
                _checksum = 0;
                
                /* start parsing the binary message */
                _state = BIN_LEN_0;
                break;
            } 
            
            /* Do we have a reasonable NMEA start sequence? */
            if( (_sent[0] == '$') && (c == 'G' || c == 'P')) {
                /* got a reasonable NMEA start sequence */
                _mode = NMEA;
                
                /* update checksum for "skipped" character */
                _checksum = c;
                
                /* mark the start point */
                _term[_tc++] = _sent + 1;
                
                /* start buffering NMEA message */
                _state = NMEA_BUFF;
                break;
            } 
            
            if( (c & 0x80) ) {
                /* this is an illegal ASCII character so we're probs in BIN */
                _mode = BINARY;
            }
        
            /* pretend to re-init */
            _sc = 1;
            _sent[0] = c;
       
             /* try again next time */
            _state = WAIT_START; 
            break;
        
        case BIN_LEN_0:
            /* validate first length byte (length < 2^10-1) */
            if( (c & (~0x3F)) ) {
                /* Invalid length, reset FSM */
                _state = RESET_FSM;
                break;
            }
            
            /* Read in the most sig. length byte*/
            _bc = ((int)c)<<8;
            
            /* read in the least sig. length byte */
            _state = BIN_LEN_1;            
            break;
        
        case BIN_LEN_1:
            /* Complete length word */
            _bc |= ((int)c);
            
            if( (_bc--) == 0){
                /* No bytes to come?? */
                _state = RESET_FSM;
                break;
            }
            
            if( (_bc > MAX_SENTENCE_LEN) ){
                /** Too many, abort! **/
                _state = RESET_FSM;
                break;
            }
            
             /* store pointer to data start */
            _term[_tc++] = _sent + _sc;
            
            /* Start buffering data */
            _state = BIN_BUFF_N;
            break;
        
        case BIN_BUFF_N:
            /** Update checksum */
            _checksum += IV(c);
            
            if( _bc-- ) {
                /* More data to come... keep going... */
                _state = BIN_BUFF_N;
                break;
            }
            
            /* No more data! Read in checksum */
            _state = BIN_CHECKSUM_0;            
            break;
        
        case BIN_CHECKSUM_0:
            /* store and verify upper byte of checksum */
            _checksum ^= IV(c)<<8;

            /* handle second checksum character */
            _state = BIN_CHECKSUM_1;
            break;
            
        case BIN_CHECKSUM_1:
            /* Store and verify lower byte of checksum */
            _checksum ^= IV(c);

            /* check end characters */
            _state = BIN_END_0;
            break;
            
        case BIN_END_0:
            if( IV(c) != 0xB0 ) {
                /* invalid end character */
                _state = RESET_FSM;
                break;
            }
            
            /* expected end character! */
            _state = BIN_END_1;
            break;
            
        case BIN_END_1:
            
            if( IV(c) != 0xB3 ){
                /* invalid end character, reset FSM */
                _state = RESET_FSM;
                break;
            }
            
            if( _checksum ){
                /* invalid checksum, reset FSM */
                _state = RESET_FSM;
                break;
            }
            
            /* record actual length =_sc - 2*start+2*len+2*check+2*stop */
            _bc = _sc - 8; 
            
            /* all good! */
            _state = BIN_VALID;
            return TRUE;
            
        case NMEA_BUFF:
            /* end of sentence ? */
            if( c == '*' ) {                
                /* mark the last term */
                _term[_tc++] = _sent + _sc;
                _term[_tc] = NULL;
                
                /* start checksum */
                _state = NMEA_CHECKSUM_0;
                break;
            } 
                
            /* check for delim, store term start position */
            if( c == ',' ) {
                /*store term start location */
                _term[_tc++] = _sent + _sc;
            }
            
            /* update checksum */
            _checksum ^= (int)(c);
            
            /* keep at it */
            _state = NMEA_BUFF;
            break;
        
        case NMEA_CHECKSUM_0:
            /* first char after * is the most significant nibble of the chksum*/
            c = _dehex(c);
            if(c & 0xF0){
                /* invalid hex! */
                _state = RESET_FSM;
                break;  
            } 
            
            /* Check upper nibble of checksum */
            _checksum -= (int)(IV(c) << 4);
            _state = NMEA_CHECKSUM_1;
            break;
        
        case NMEA_CHECKSUM_1:
            /* Null terminate string. */
            _sent[_sc] = 0; 
            
            
            /* second char after * is the least sig. nibble of checksum */
            c = _dehex(c);
            
            if(c & 0xF0){
                /* Invalid hex! */
                _state = RESET_FSM;
                break;
            } 

            _checksum -= IV(c);
            
            if(_checksum){
                /* invalid checksum */
                _state = RESET_FSM;
                break;
            }
            
            _bc = _sc;
            
            _state = NMEA_VALID;            
            return TRUE;

        default:
            _state = RESET_FSM;
    }
    
    return FALSE;
}

int SiRF::_break_sent(void){
    if(_state != NMEA_VALID) return 0;
    if(!_sc)  return 0;

#ifdef _KEEP_SENTENCE_
    /* If user has requested a copy of sentence, do it now */
    memcpy(_sent2, _sent, _sc);
#endif
    _sc = 0;

    int i = 0;
    /* Split terms into valid strings */
    while(i < _tc) *(_term[i++]-1) = 0;
    /* Invalidate remaining terms */
    while(i < MAX_NUM_TERMS) _term[i++] = NULL;

    return _tc;
}

/**************************** File-Private Functions **********************/

/* Tuned routine to parse NMEA longitude. Uses integer arithmetic as long as 
 possible. 
Format = DDSS.ssss 
    D = Whole degree
    S = Whole seconds
    s = fractional seconds
*/
static fp_t _lat2dd(char* s){
    #define char2dec(s, i, l)     (i < l ? (s[i]-'0') : 0)
    /* check len and format */
    int l = strlen(s);
    if(l < 5 || s[4] != '.') return (-1.0);
    
    /* Integer multiplication is faster, and immune to rounding errors */
    int32_t deg =                       char2dec(s, 0, l);         // D
    deg =           times10plus(deg,    char2dec(s, 1, l)  );      // D
    
    
    /* unrolled loop */        
    int32_t sec =                       char2dec(s, 2, l)   ;      // S
    sec =           times10plus(sec,    char2dec(s, 3, l)  );      // S
                                                                   // . 
    sec =           times10plus(sec,    char2dec(s, 5, l)  );      // s
    sec =           times10plus(sec,    char2dec(s, 6, l)  );      // s
    sec =           times10plus(sec,    char2dec(s, 7, l)  );      // s
    sec =           times10plus(sec,    char2dec(s, 8, l)  );      // s
                    
    /* do only one division, to convert from seconds*1000 to decimal degs */
    fp_t rv = (fp_t) sec;
    rv /= (10*60*1000.0);
    
    /* add integer degrees */
    rv += deg;
    return rv;
    #undef char2dec
}

/* Tuned routine to parse NMEA longitude. Uses integer arithmetic as long as 
 possible. 
Format = DDDSS.ssss 
    D = Whole degree
    S = Whole seconds
    s = fractional seconds
*/
static fp_t _lon2dd(char* s){
    #define char2dec(s, i, l)     (i < l ? (s[i]-'0') : 0)
    /* check len and format */
    int l = strlen(s);
    if(l < 6 || s[5] != '.') return (-1.0);
    
    /* Integer multiplication is faster, and immune to rounding errors */
    int32_t deg =                       char2dec(s, 0, l);         // D
    deg =           times10plus(deg,    char2dec(s, 1, l)  );      // D
    deg =           times10plus(deg,    char2dec(s, 2, l)  );      // D
    
    
    /* unrolled loop */        
    int32_t sec =                       char2dec(s, 3, l)   ;      // S
    sec =           times10plus(sec,    char2dec(s, 4, l)  );      // S
                                                                // . 
    sec =           times10plus(sec,    char2dec(s, 6, l)  );      // s
    sec =           times10plus(sec,    char2dec(s, 7, l)  );      // s
    sec =           times10plus(sec,    char2dec(s, 8, l)  );      // s
    sec =           times10plus(sec,    char2dec(s, 9, l)  );      // s
    
    
    /* do only one division, to convert from seconds*1000 to decimal degs */
    fp_t rv = (fp_t) sec;
    rv /= (10*60*1000.0);
    
    /* add integer degrees */
    rv += deg;
    return rv;
    #undef char2dec
}


/* Attempts to parse two decimal digits from a string pointed to by sp */
static char _bcd2dec2(char** sp)
{
    char rv = 0, c;
    if( *(*sp)  && ((c = _dehex(*((*sp)++))) < 0x0A) ) rv += c;
    else return rv; // This is *(*Sparta++)?
    if( *(*sp)  && ((c = _dehex(*((*sp)++))) < 0x0A) ) rv = (10*rv) + c;
    return rv;
}

/* probably faster than branching?  */
static char _dehex(char c)
{
    switch(c){
        case '0': return 0x00; case '1': return 0x01;
        case '2': return 0x02; case '3': return 0x03;
        case '4': return 0x04; case '5': return 0x05;
        case '6': return 0x06; case '7': return 0x07;
        case '8': return 0x08; case '9': return 0x09;
        case 'A': return 0x0A; case 'B': return 0x0B;
        case 'C': return 0x0C; case 'D': return 0x0D;
        case 'E': return 0x0E; case 'F': return 0x0F;
        default:  return 0xF0;
    }
}

