/*******************************************************************************
* sirf.h <=> sirf.cpp: NMEA 0183 & SiRF Binary Protocol library for Arduino    *
* Copyright (c) 2012 Jacob Dilles <jdilles@gmu.edu>        All right reserved. *
*                                                                              *
* This software has been released under The MIT License (MIT):                 *
*                                                                              *
* Permission is hereby granted, free of charge, to any person obtaining a copy *
* of this software and associated documentation files (the "Software"), to     *
* deal in the Software without restriction, including without limitation the   *
* rights to use, copy, modify, merge, publish, distribute, sublicense, and/or  *
* sell copies of the Software, and to permit persons to whom the Software is   *
* furnished to do so, subject to the following conditions:                     *
*                                                                              *
* (1) The above copyright notice and this permission notice shall be included  *
* in all copies or substantial portions of the Software.                       *
*                                                                              *
* (2) THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  *
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       *
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING      *
* FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS *
* IN THE SOFTWARE.                                                             *
*                                                                              *
* Notice: The author has no affiliation with SiRF Technology, Inc. or CSR plc. *
*                                                                              *
* Many thanks to Maarten Lamers for his NMEA library:                          *
*       http://www.maartenlamers.com/nmea/                                     *
* Which inspired this project and got it rolling in the right direction.       *
*******************************************************************************/    

#ifndef sirf_h
#define sirf_h

#include "Arduino.h"
#include <stdint.h>


#define MAX_SENTENCE_LEN        (128)
#define MAX_TERM_LEN            (16)
#define MAX_NUM_TERMS           (32)      
//#define _KEEP_SENTENCE_

#define fp_t                    float 


struct st_gprmc {
	char status;
	char utc_year; char utc_month; char utc_day;
	char utc_hour; char utc_min; char utc_sec; 
	fp_t lat;
	fp_t lon;
	fp_t speed;
	fp_t angle;
};

typedef struct st_gprmc GPRMC;

enum mode_t {
    INVALID = 0,
    BINARY  = 1,
    NMEA    = 2
};



fp_t distance(struct st_gprmc * d1, struct st_gprmc * d2);
fp_t bearing(struct st_gprmc * d1, struct st_gprmc * d2); 

/* The actual object */
class SiRF
{
  public:
    	SiRF();
    	/* 
    	start the serial stuff and set the baud rate
    	    args: baud   - 4800 *highly* recommended
    	*/
    	void begin(const HardwareSerial * stream, unsigned long baud);
    	
    	/* 
    	Send a NMEA String (with fun Arduino overrides). Calculates checksum etc
    	    Takes:  Most types of strings 
    	    Returns: True if we are in NMEA mode and the string sent
    	*/
    	int sendNMEA(char*);
    	int sendNMEA(const char*);
    	int sendNMEA(const String &s);
    	int sendNMEA(const __FlashStringHelper *ifsh);
    	
    	/* 
        Sends an SiRF Binary Protocol String, calcuates checksums, etc.. 
            params:     buffp - pointer to buffer
                        lenp - length to send
                        
            returns:    boolean error
            
            reference: SiRF Binary Protocol Reference Manual 
                       Revision 2.4 November 2008 
        */
    	int sendBIN(char* buffp, int16_t lenp);
    	int sendBIN(const char* buffp, int16_t lenp);
    	
    	/* 
        Sends an Hex-encoded SiRF Binary Protocol String, calcuates checksums, etc.. 
            params:     str - hex encoded string
                        
            returns:    boolean error
            
            reference: SiRF Binary Protocol Reference Manual 
                       Revision 2.4 November 2008 
        */    	
    	int sendHEX(const char* str);
    	
        /* 
          read the serial port until it is out of characters or we have a valid sentence
          returns:  NMEA, BINARY if valid sentance recieved
                    0 otherwise
        */
        mode_t receive(void);
    
        /* 
        Parse a GPRMC sentence into the specified struct 
            args:       d - pointer to data struct
            returns:    TRUE if parsed OK, FALSE on error 
        */
        int     parseGPRMC(GPRMC * d); 
    
        /*
            sentence() Return the last sentence recieived
            sentenceLength() returns the length of the sentence  
        */
        char*   sentence();
        int     sentenceLength();
        
        /* 
            terms() returns the number of terms in the last sentence,
            term(t) returns the t-th term
        */
        int     terms();
        char*	term(int t);    
    
        
  private:
		
		/* FSM to handle one char of input data */ 
		int _decode(char c);
		
		/* break the sentence up */
		int  _break_sent(void);
	
	    char*  _term[MAX_NUM_TERMS];
	    char   _sent[MAX_SENTENCE_LEN  +  8];
        
#ifdef _KEEP_SENTENCE_
        /* if we keep the sentence whole, the copy goes here*/
        char  _sent2[MAX_SENTENCE_LEN + 8];
#endif
	    
		int16_t   _tc;       /* term  # counter */
        int16_t   _cc;       /* term character counter */
        int16_t   _sc;       /* sentence counter */
        int16_t   _bc;       /* binary counter */
        uint16_t  _checksum ; /* checksum */
        
        HardwareSerial * _serial;
	uint16_t _baud;
        
        mode_t _mode;     /* binary or NMEA */
        
        enum state_t {
            RESET_FSM, WAIT_START,
            
            BIN_LEN_0, BIN_LEN_1, BIN_BUFF_N, BIN_CHECKSUM_0, BIN_CHECKSUM_1,
            BIN_END_0, BIN_END_1, BIN_VALID,
            
            NMEA_BUFF, NMEA_CHECKSUM_0, NMEA_CHECKSUM_1, NMEA_VALID
        } _state;
};


/*Extras we need */
#ifndef NULL
#define NULL (0)
#endif
#ifndef NAN
#define NAN ((0.0)/(0.0))
#endif
#ifndef TRUE
#define TRUE (1)
#endif
#ifndef FALSE
#define FALSE (0)
#endif


#endif
